﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2018 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis {
  [Item("Data Analysis Transformation Model", "A model that was transformed back to match the original variables after the training was performed on transformed variables.")]
  [StorableClass]
  public abstract class DataAnalysisTransformationModel : DataAnalysisModel, IDataAnalysisTransformationModel {

    [Storable]
    public IDataAnalysisModel OriginalModel { get; protected set; }

    [Storable]
    public ReadOnlyItemList<IDataAnalysisTransformation> InputTransformations { get; protected set; }

    [Storable]
    public ReadOnlyItemList<IDataAnalysisTransformation> TargetTransformations { get; protected set; }

    // Usually, the TargetVariable is usually only implemented for Regression and Classification.
    // However, we implement it in the base class for reducing code duplication and to avoid quasi-identical views.
    [Storable]
    private string targetVariable;
    public string TargetVariable {
      get { return targetVariable; }
      set {
        if (string.IsNullOrEmpty(value) || targetVariable == value) return;
        targetVariable = value;
        OnTargetVariableChanged(this, EventArgs.Empty);
      }
    }

    public override IEnumerable<string> VariablesUsedForPrediction {
      get { return OriginalModel.VariablesUsedForPrediction; /* TODO: reduce extend-inputs */}
    }

    #region Constructor, Cloning & Persistence
    protected DataAnalysisTransformationModel(IDataAnalysisModel originalModel, IEnumerable<IDataAnalysisTransformation> transformations)
      : base(originalModel.Name) {
      OriginalModel = originalModel;
      var transitiveInputs = CalculateTransitiveVariables(originalModel.VariablesUsedForPrediction, transformations);
      InputTransformations = new ItemList<IDataAnalysisTransformation>(transformations.Where(t => transitiveInputs.Contains(t.OriginalVariable))).AsReadOnly();
      TargetTransformations = new ReadOnlyItemList<IDataAnalysisTransformation>();
    }

    protected DataAnalysisTransformationModel(DataAnalysisTransformationModel original, Cloner cloner)
      : base(original, cloner) {
      OriginalModel = cloner.Clone(original.OriginalModel);
      InputTransformations = cloner.Clone(original.InputTransformations);
      TargetTransformations = cloner.Clone(original.TargetTransformations);
      targetVariable = original.targetVariable;
    }

    [StorableConstructor]
    protected DataAnalysisTransformationModel(bool deserializing)
      : base(deserializing) { }
    #endregion

    // extended -> include originals
    public static ISet<string> CalculateTransitiveVariables(IEnumerable<string> inputVariables, IEnumerable<IDataAnalysisTransformation> transformations) {
      var transitiveInputs = new HashSet<string>(inputVariables);

      foreach (var transformation in transformations.Reverse()) {
        if (transitiveInputs.Contains(transformation.TransformedVariable)) {
          transitiveInputs.Add(transformation.OriginalVariable);
        }
      }

      return transitiveInputs;
    }
    // originals => include extended
    public static IEnumerable<string> ExtendInputVariables(IEnumerable<string> oldInputVariables, IEnumerable<IDataAnalysisTransformation> transformations) {
      var inputs = new HashSet<string>(oldInputVariables);

      foreach (var transformation in transformations) {
        if (inputs.Contains(transformation.OriginalVariable))
          inputs.Add(transformation.TransformedVariable);
      }

      return inputs;
    }
    [Obsolete]
    public static IEnumerable<string> RemoveVirtualVariables(IEnumerable<string> variables, IEnumerable<IDataAnalysisTransformation> transformations) {
      var remainingVariables = new HashSet<string>(variables);

      var transformationsStack = new Stack<IDataAnalysisTransformation>(transformations);

      while (transformationsStack.Any()) {
        var transformation = transformationsStack.Pop();


        bool transformedVariablePending = transformationsStack.Any(x => x.OriginalVariable == transformation.TransformedVariable);
        if (!transformedVariablePending)
          remainingVariables.Remove(transformation.TransformedVariable);
      }

      return remainingVariables;
    }

    public static IDataset Transform(IDataset dataset, IEnumerable<IDataAnalysisTransformation> transformations) {
      var modifiableDataset = ((Dataset)dataset).ToModifiable();

      foreach (var transformation in transformations) {
        var trans = (ITransformation<double>)transformation.Transformation;

        var originalData = modifiableDataset.GetDoubleValues(transformation.OriginalVariable);
        if (!trans.Check(originalData, out string errorMessage))
          throw new InvalidOperationException($"Cannot estimate Values, Transformation is invalid: {errorMessage}");

        var transformedData = trans.Apply(originalData).ToList();
        if (modifiableDataset.VariableNames.Contains(transformation.TransformedVariable))
          modifiableDataset.ReplaceVariable(transformation.TransformedVariable, transformedData);
        else
          modifiableDataset.AddVariable(transformation.TransformedVariable, transformedData);
      }

      return modifiableDataset;
    }

    #region Events
    public event EventHandler TargetVariableChanged;
    private void OnTargetVariableChanged(object sender, EventArgs args) {
      var changed = TargetVariableChanged;
      if (changed != null)
        changed(sender, args);
    }
    #endregion
  }
}